/*****************************************************************
**
**	@(#) misc.c -- helper functions 
**
**	Copyright (c) Jan 2005, Holger Zuleger HZNET. All rights reserved.
**
**	This software is open source.
**
*****************************************************************/
# include <stdio.h>
# include <string.h>
# include <stdlib.h>
# include <unistd.h>	/* for link(), unlink() */
# include <ctype.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <time.h>
# include <utime.h>
# include <assert.h>
# include <errno.h>
# include <fcntl.h>
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
# include "debug.h"
#define extern
# include "misc.h"
#undef extern
# include "spmake.h"

# define	TAINTEDCHARS	"^=`$@&"

extern	const	char	*progname;

static	int	normdir (char *dir);

/*****************************************************************
**	copyfile (fromfile, tofile)
**	copy fromfile into tofile.
*****************************************************************/
int	copyfile (const char *fromfile, const char *tofile)
{
	FILE	*infp;
	FILE	*outfp;
	int	c;

	/* fprintf (stderr, "copyfile (%s, %s)\n", fromfile, tofile); */
	if ( (infp = fopen (fromfile, "r")) == NULL )
		return -1;
	if ( (outfp = fopen (tofile, "w")) == NULL )
	{
		fclose (infp);
		return -2;
	}
	while ( (c = getc (infp)) != EOF ) 
		putc (c, outfp);

	fclose (infp);
	fclose (outfp);

	return 0;
}


/*****************************************************************
**	str_delspace (s)
**	Remove in string 's' all white space char 
*****************************************************************/
char	*str_delspace (char *s)
{
	char	*start;
	char	*p;

	if ( !s )	/* no string present ? */
		return NULL;

	start = s;
	for ( p = s; *p; p++ )
		if ( !isspace (*p) )
			*s++ = *p;	/* copy each nonspace */

	*s = '\0';	/* terminate string */

	return start;
}

/*****************************************************************
**	in_strarr (str, arr, cnt)
**	check if string array 'arr' contains the string 'str'
**	return 1 if true or 'arr' or 'str' is empty, otherwise 0
*****************************************************************/
int	in_strarr (const char *str, char *const arr[], int cnt)
{
	if ( arr == NULL || cnt <= 0 )
		return 1;

	if ( str == NULL || *str == '\0' )
		return 0;

	while ( --cnt >= 0 )
		if ( strcmp (str, arr[cnt]) == 0 )
			return 1;

	return 0;
}

/*****************************************************************
**	str_untaint (s)
**	Remove in string 's' all TAINTED chars
*****************************************************************/
char	*str_untaint (char *str)
{
	char	*p;

	assert (str != NULL);

	for ( p = str; *p; p++ )
		if ( strchr (TAINTEDCHARS, *p) )
			*p = ' ';
	return str;
}

/*****************************************************************
**	str_chop (str, c)
**	delete all occurrences of char 'c' at the end of string 's'
*****************************************************************/
char	*str_chop (char *str, char c)
{
	int	len;

	assert (str != NULL);

	len = strlen (str) - 1;
	while ( len >= 0 && str[len] == c )
		str[len--] = '\0';

	return str;
}

/*****************************************************************
**	issubdir (parent, subdir)
*****************************************************************/
int	issubdir (const char *pdir, const char *sdir)
{
	char	workingdir[MAXPATHLEN];
	char	parent[MAXPATHLEN];
	char	child[MAXPATHLEN];
	int	len;

	if ( *pdir != '/' || *sdir != '/' )
		if ( getcwd (workingdir, MAXPATHLEN) == NULL )
			fatal ("Can't get current working directory\n");

	if ( *pdir == '/' )
		snprintf (parent, MAXPATHLEN, "%s", pdir);
	else
		snprintf (parent, MAXPATHLEN, "%s/%s", workingdir, pdir);
	len = normdir (parent);

	if ( *sdir == '/' )
		snprintf (child, MAXPATHLEN, "%s", sdir);
	else
		snprintf (child, MAXPATHLEN, "%s/%s", workingdir, sdir);
	normdir (child);

	return ( strncmp (parent, child, len) == 0 );
}

/*****************************************************************
**	int	normdir (path)
**	remove all '.' and '..' path components in path
**	path *must* be an absolute path thus starts with an '/'
**	returns the new length of path
*****************************************************************/
static	int	normdir (char *dir)
{
	char	*a;
	char	*b;
	char	*lastdelim;

	assert ( dir != NULL );
	assert ( *dir == '/' );

	lastdelim = a = b = dir;
	while ( *b )
	{
		if ( b[0] == '/' )	
		{
			if ( b[1] == '/' )			/* skip '//' */
				b += 1;	
			else if ( b[1] == '.' && (b[2] == '/' || b[2] == '\0') )	/* skip '/.' */
				b += 2;	
			else if ( b[1] == '.' && b[2] == '.' && (b[3] == '/' || b[3] == '\0') )	/* handle '/..' */
			{
				b += 3;
				a = lastdelim;		/* set 'a' back to the last '/' */
			}
			else
			{
				lastdelim = a;		/* save position of '/' */
				*a++ = *b++;	/* copy */
			}
		}
		else
			*a++ = *b++;	/* copy */
	}
	*a = '\0';

	return (a - dir);
}

/*****************************************************************
**	splitpath (path, pathsize, filename)
**	if filename is build of "path/file" then copy filename to path
**	and split of the filename part.
**	return pointer to filename part in path or NULL if path is too
**	small to hold "path+filename"
*****************************************************************/
const	char	*splitpath (char *path, size_t psize, const char *filename)
{
	char 	*p;

	if ( !path )
		return NULL;

	*path = '\0';
	if ( !filename )
		return filename;

	if ( (p = strrchr (filename, '/')) )	/* file arg contains path ? */
	{
		if ( strlen (filename) + 1 > psize )
			return filename;

		strcpy (path, filename);	/* copy whole filename to path */
		path[p-filename] = '\0';	/* split of the file part */
		filename = ++p;
	}
	return filename;
}

/*****************************************************************
**	pathname (path, size, dir, file, ext)
**	Concatenate 'dir', 'file' and 'ext' (if not null) to build
**	a pathname, and store the result in the character array
**	with length 'size' pointed to by 'path'.
**	if 'replext' is given then file contains alread and extension 'ext',
**	and this one will be replaced by 'replext'.
*****************************************************************/
char	*pathname (char *path, size_t size, const char *dir, const char *file, const char *ext, const char *repl)
{
	int	len;
	int	extlen;
	int	replen;

	if ( path == NULL || file == NULL )
		return path;

	len = strlen (file) + 1;
	if ( dir )
		len += strlen (dir);

	extlen = replen = 0;
	if ( ext )
	{
		extlen = strlen (ext);
		if ( repl )
		{
			replen = strlen (repl);
			len += strlen (repl);
		}
		else
			len += extlen;
	}

	if ( len > size )	/* doesn't fit ? */
		return path;

	*path = '\0';
	len = 0;
	if ( dir && *dir )
	{
		len = sprintf (path, "%s", dir);
		if ( path[len-1] != '/' )
		{
			path[len++] = '/';
			path[len] = '\0';
		}
	}
	len += sprintf (path+len, "%s", file);
	if ( ext )
	{
		if ( repl == NULL )		/* if only 'ext' is given ? */
			len += sprintf (path+len, "%s", ext);
		else			/* otherwise replace 'ext' with 'repl' */
		{
			len -= extlen;
			if ( strcmp (path+len, ext) == 0 )		/* if ext is already there */
				len += sprintf (path+len, "%s", repl);
		}
	}
	return path;
}

/*****************************************************************
**	is_directory (name)
**	Check if the given pathname 'name' exists and is a directory.
**	returns 0 | 1
*****************************************************************/
int	is_directory (const char *name)
{
	struct	stat	st;

	if ( !name || !*name )	
		return 0;
	
	return ( stat (name, &st) == 0 && S_ISDIR (st.st_mode) );
}

/*****************************************************************
**	fileexist (name)
**	Check if a file with the given pathname 'name' exists.
**	returns 0 | 1
*****************************************************************/
int	fileexist (const char *name)
{
	struct	stat	st;
	return ( stat (name, &st) == 0 && S_ISREG (st.st_mode) );
}

/*****************************************************************
**	filesize (name)
**	return the size of the file with the given pathname 'name'.
**	returns -1 if the file not exist 
*****************************************************************/
size_t	filesize (const char *name)
{
	struct	stat	st;
	if  ( stat (name, &st) == -1 )
		return -1L;
	return ( st.st_size );
}

/*****************************************************************
**	has_fileext (name, ext)
**	Check if the given file has extension "ext". 
**	Returns 0 | 1
*****************************************************************/
int	has_fileext (const char *name, const char *ext)
{
	int	len;
	int	extlen;

	assert ( name != NULL );

	dbg_val2 ("has_filext (%s, %s)\n", name, ext);
	if ( ext == NULL || *ext == '\0' )
		return 1;

	extlen = strlen (ext);
	len = strlen (name);
	dbg_val ("\tlen_name: %d\n", len);
	dbg_val ("\tlen_ext : %d\n", extlen);
	if ( len > extlen && strcmp (name + (len - extlen), ext) == 0 ) 
		return 1;

	return 0;
}

/*****************************************************************
**	is_dotfilename (name)
**	Check if the given pathname 'name' looks like "." or "..".
**	Returns 0 | 1
*****************************************************************/
int	is_dotfilename (const char *name)
{
	if ( name && (
	     (name[0] == '.' && name[1] == '\0') || 
	     (name[0] == '.' && name[1] == '.' && name[2] == '\0')) )
		return 1;

	return 0;
}

/*****************************************************************
**	touch (name, sec)
**	Set the modification time of the given pathname 'fname' to
**	'sec'.	Returns 0 on success.
*****************************************************************/
int	touch (const char *fname, time_t sec)
{
	struct	utimbuf	utb;

	utb.actime = utb.modtime = sec;
	return utime (fname, &utb);
}

#if 0
/*****************************************************************
**	linkfile (fromfile, tofile)
*****************************************************************/
int	linkfile (const char *fromfile, const char *tofile)
{
	int	ret;

	/* fprintf (stderr, "linkfile (%s, %s)\n", fromfile, tofile); */
	if ( (ret = link (fromfile, tofile)) == -1 && errno == EEXIST )
		if ( unlink (tofile) == 0 )
			ret = link (fromfile, tofile);

	return ret;
}
#endif

/*****************************************************************
**	file_age (fname)
*****************************************************************/
int	file_age (const char *fname)
{
	time_t	curr = time (NULL);
	time_t	mtime = file_mtime (fname);

	return curr - mtime;
}

/*****************************************************************
**	file_mtime (fname)
*****************************************************************/
time_t	file_mtime (const char *fname)
{
	struct	stat	st;

	if ( stat (fname, &st) < 0 )
		return 0;
	return st.st_mtime;
}

#if 0
/*****************************************************************
**	is_exec_ok (prog)
**	Check if we are running as root or if the file owner of
**	"prog" do not match the current user or the file permissions
**	allows file modification for others then the owner.
**	The same condition will be checked for the group ownership.
**	return 1 if the execution of the command "prog" will not
**	open a big security whole, 0 otherwise
*****************************************************************/
int	is_exec_ok (const char *prog)
{
	uid_t	curr_uid;
	struct	stat	st;

	if ( stat (prog, &st) < 0 )
		return 0;

	curr_uid = getuid ();
	if ( curr_uid == 0 )			/* don't run the cmd if we are root */
		return 0;

	/* if the file owner and the current user matches and */
	/* the file mode is not writable except for the owner, we are save */
	if ( curr_uid == st.st_uid && (st.st_mode & (S_IWGRP | S_IWOTH)) == 0 )
		return 1;

	/* if the file group and the current group matches and */
	/* the file mode is not writable except for the group, we are also save */
	if ( getgid() != st.st_gid && (st.st_mode & (S_IWUSR | S_IWOTH)) == 0 )
		return 1;

	return 0;
}
#endif

/*****************************************************************
**	fatal (fmt, ...)
*****************************************************************/
void	fatal (char *fmt, ...)
{
        va_list ap;

        va_start(ap, fmt);
        if ( progname )
		fprintf (stderr, "%s: ", progname);
        vfprintf (stderr, fmt, ap);
        va_end(ap);
        exit (127);
}

/*****************************************************************
**	error (fmt, ...)
*****************************************************************/
void	error (char *fmt, ...)
{
        va_list ap;

        va_start(ap, fmt);
        vfprintf (stderr, fmt, ap);
        va_end(ap);
}

/*****************************************************************
**	logmesg (fmt, ...)
*****************************************************************/
void logmesg (char *fmt, ...)
{
        va_list ap;

#if defined (LOG_WITH_PROGNAME) && LOG_WITH_PROGNAME
        fprintf (stdout, "%s: ", progname);
#endif
        va_start(ap, fmt);
        vfprintf (stdout, fmt, ap);
        va_end(ap);
}

/*****************************************************************
**	verbmesg (verblvl, conf, fmt, ...)
*****************************************************************/
void	verbmesg (int verblvl, char *fmt, ...)
{
	char	str[511+1];
        va_list ap;

	str[0] = '\0';
	va_start(ap, fmt);
	vsnprintf (str, sizeof (str), fmt, ap);
	va_end(ap);

	//fprintf (stderr, "verbmesg (%d stdout=%d filelog=%d str = :%s:\n", verblvl, conf->verbosity, conf->verboselog, str);
	if ( verblvl <= verbose )	/* check if we have to print this to stdout */
		logmesg (str);
}


/*****************************************************************
**	logflush ()
*****************************************************************/
void logflush ()
{
        fflush (stdout);
}

/*****************************************************************
**	time2str (sec, precison)
**	sec is seconds since 1.1.1970
**	precison is currently either 's' (for seconds) or 'm' (minutes)
*****************************************************************/
char	*time2str (time_t sec, int precision)
{
	struct	tm	*t;
	static	char	timestr[31+1];	/* 27+1 should be enough */
#if defined(HAVE_STRFTIME) && HAVE_STRFTIME
	char	tformat[127+1];

	timestr[0] = '\0';
	if ( sec <= 0L )
		return timestr;
	t = localtime (&sec);
	if ( precision == 's' )
		strcpy (tformat, "%b %d %Y %T");
	else
		strcpy (tformat, "%b %d %Y %R");
# if PRINT_TIMEZONE
	strcat (tformat, " %z");
# endif
	strftime (timestr, sizeof (timestr), tformat, t);

#else	/* no strftime available */
	static	char	*mstr[] = {
			"Jan", "Feb", "Mar", "Apr", "May", "Jun",
			"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
	};

	timestr[0] = '\0';
	if ( sec <= 0L )
		return timestr;
	t = localtime (&sec);
# if PRINT_TIMEZONE
	{
	int	h,	s;

	s = abs (t->tm_gmtoff);
	h = t->tm_gmtoff / 3600;
	s = t->tm_gmtoff % 3600;
	if ( precision == 's' )
		snprintf (timestr, sizeof (timestr), "%s %2d %4d %02d:%02d:%02d %c%02d%02d",
			mstr[t->tm_mon], t->tm_mday, t->tm_year + 1900, 
			t->tm_hour, t->tm_min, t->tm_sec,
			t->tm_gmtoff < 0 ? '-': '+',
			h, s);
	else
		snprintf (timestr, sizeof (timestr), "%s %2d %4d %02d:%02d %c%02d%02d",
			mstr[t->tm_mon], t->tm_mday, t->tm_year + 1900, 
			t->tm_hour, t->tm_min, 
			t->tm_gmtoff < 0 ? '-': '+',
			h, s);
	}
# else
	if ( precision == 's' )
		snprintf (timestr, sizeof (timestr), "%s %2d %4d %02d:%02d:%02d",
			mstr[t->tm_mon], t->tm_mday, t->tm_year + 1900, 
			t->tm_hour, t->tm_min, t->tm_sec);
	else
		snprintf (timestr, sizeof (timestr), "%s %2d %4d %02d:%02d",
			mstr[t->tm_mon], t->tm_mday, t->tm_year + 1900, 
			t->tm_hour, t->tm_min);
# endif
#endif

	return timestr;
}

/*****************************************************************
**	time2isostr (str, strlen, sec, precison)
**	sec is seconds since 1.1.1970
**	precison is currently either 's' (for seconds), 'm' (minutes)
**	'h' (hours) or 'd' for days 
*****************************************************************/
int	time2isostr (char *timestr, int tstrlen, time_t sec, int precision)
{
	struct	tm	*t;
	int	len;

	assert ( timestr != NULL );
	assert ( tstrlen >= 1 );

	if ( sec == 0L )
	{
		*timestr = '\0';
		return 0;
	}

	t = localtime (&sec);
	if ( precision >= DAYSEC || precision == 'd' )
		len = snprintf (timestr, tstrlen, "%4d%02d%02d",
			t->tm_year + 1900, t->tm_mon+1, t->tm_mday);
	else if ( precision >= HOURSEC || precision == 'h' )
		len = snprintf (timestr, tstrlen, "%4d%02d%02dT%02d",
			t->tm_year + 1900, t->tm_mon+1, t->tm_mday,
			t->tm_hour);
	else if ( precision >= MINSEC || precision == 'm' )
		len = snprintf (timestr, tstrlen, "%4d%02d%02dT%02d%02d",
			t->tm_year + 1900, t->tm_mon+1, t->tm_mday,
			t->tm_hour, t->tm_min);
	else
		len = snprintf (timestr, tstrlen, "%4d%02d%02dT%02d%02d%02d",
			t->tm_year + 1900, t->tm_mon+1, t->tm_mday,
			t->tm_hour, t->tm_min, t->tm_sec);

	return len;
}

/*****************************************************************
**	agefromstr (str)
*****************************************************************/
int	agefromstr (const char *str)
{
	int	age;

	assert ( str != NULL );

	age = atoi (str);

	while ( isspace (*str) )
		str++;
	while ( isdigit (*str) )
		str++;
	switch ( tolower (*str) )
	{
	case 'd':	age *= DAYSEC;	break;
	case 'h':	age *= HOURSEC;	break;
	case 'm':	age *= MINSEC;	break;
	}

	return age;
}

#if 0
/*****************************************************************
**	age2str (sec)
**	!!Attention: This function is not reentrant 
*****************************************************************/
char	*age2str (time_t sec)
{
	static	char	str[20+1];	/* "2y51w6d23h50m55s" == 16+1 chars */
	int	len;
	int	strsize = sizeof (str);

	len = 0;
# if PRINT_AGE_WITH_YEAR
	if ( sec / (YEARSEC) > 0 )
	{
		len += snprintf (str+len, strsize - len, "%1luy", sec / YEARSEC );
		sec %= (YEARSEC);
	}
	else
		len += snprintf (str+len, strsize - len, "  ");
# endif
	if ( sec / WEEKSEC > 0 )
	{
		len += snprintf (str+len, strsize - len, "%2luw", (ulong) sec / WEEKSEC );
		sec %= WEEKSEC;
	}
	else
		len += snprintf (str+len, strsize - len, "   ");
	if ( sec / DAYSEC > 0 )
	{
		len += snprintf (str+len, strsize - len, "%2lud", sec / (ulong)DAYSEC);
		sec %= DAYSEC;
	}
	else
		len += snprintf (str+len, strsize - len, "   ");
	if ( sec / HOURSEC > 0 )
	{
		len += snprintf (str+len, strsize - len, "%2luh", sec / (ulong)HOURSEC);
		sec %= HOURSEC;
	}
	else
		len += snprintf (str+len, strsize - len, "   ");
	if ( sec / MINSEC > 0 )
	{
		len += snprintf (str+len, strsize - len, "%2lum", sec / (ulong)MINSEC);
		sec %= MINSEC;
	}
	else
		len += snprintf (str+len, strsize - len, "   ");
	if ( sec > 0 )
		snprintf (str+len, strsize - len, "%2lus", (ulong) sec);
	else
		len += snprintf (str+len, strsize - len, "   ");

	return str;
}
#endif

#ifdef TEST
int	verbose;
const	char	*progname;
int	main (int argc, char *argv[])
{
	char	path[MAXPATHLEN];
	int	len;

	strcpy (path, argv[1]);

	printf ("o123456789o123456789o123456789\n"),
	puts (path);
	len = normdir (path);
	printf ("%-30s %d\n", path, len);
}
#endif
