/*****************************************************************
**
**	@(#) spmake.c	(c) Jul 2011 by hznet
**
**	spmake is a simple pattern based make program
**	It is simple because it can only create one level targets,
**	and it is pattern based so there is no need to define each
**	target in a config file
**		
**	Copyright (c) 2014, Holger Zuleger HZNET. All rights reserved.
**	This software is open source.
**
**	Redistribution and use in source and binary forms, with or without
**	modification, are permitted provided that the following conditions
**	are met:
**
**	Redistribution of source code must retain the above copyright notice,
**	this list of conditions and the following disclaimer.
**
**	Redistribution's in binary form must reproduce the above copyright notice,
**	this list of conditions and the following disclaimer in the documentation
**	and/or other materials provided with the distribution.
**
**	Neither the name of Holger Zuleger, or the name HZNET nor the names of its
**	contributors may be used to endorse or promote products derived from this
**	software without specific prior written permission.
**
**	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
**	"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
**	TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
**	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
**	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
**	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
**	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
**	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
**	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
**	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
**	POSSIBILITY OF SUCH DAMAGE.
**
*****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>	/* time() */
#include <unistd.h>
#include <ctype.h>
#include <getopt.h>	/* getopt() */
#include <dirent.h>
#include <assert.h>	/* assert() */
#include <sys/stat.h>	/* mkdir() */
#include <sys/types.h>	/* mkdir() */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define extern
# include "spmake.h"
#undef extern
# include "rule.h"
# include "misc.h"
# include "debug.h"

# define	short_options	"hVceflnrvd:s:t:T::p:X:"
#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
static struct option long_options[] = {
	{"create",		no_argument, NULL, 'c'},
	{"exit",		no_argument, NULL, 'e'},
	{"force",		no_argument, NULL, 'f'},
	{"list",		no_argument, NULL, 'l'},
	{"noexec",		no_argument, NULL, 'n'},
	{"recursive",		no_argument, NULL, 'r'},
	{"verbose",		no_argument, NULL, 'v'},

	{"diff",		required_argument, NULL, 'd'},
	{"srcdir",		required_argument, NULL, 's'},
	{"targetdir",		required_argument, NULL, 't'},
	{"targetfile",		optional_argument, NULL, 'T'},
	{"pattern",		required_argument, NULL, 'p'},
	{"root-exec-allowed",	no_argument, NULL, 02 },
	{"version",		no_argument, NULL, 'V'},
	{"help",		no_argument, NULL, 'h'},
	{0, 0, 0, 0}
};
#endif


static	void	usage (const char *mesg, const char *par);
static	int	setup_rules (void);
static	int	runcmd (char *cmd, const char *src, time_t s_mtime, const char *tgt, time_t t_mtime);
static	char	*commandstr (const char *cmd, const char *src, time_t sdate, const char *tgt, time_t tdate);
static	int	parsedir_rm (const char *s_dir, const char *t_dir, rule_t *rule);
static	int	parsedir_sm (char *buf, int bufsize, const char *s_dir, time_t ttime, rule_t *rule);


const	char	*progname;
extern	int	optind;


int	create = 0;
int	diff = 0;
int	listrules = 0;
int	rootexec = 0;
int	noexec = 0;
int	force = 0;
int	singletargetmode = 0;
int	optionp = 0;
int	exitmode = 0;
int	exitvalue = 0;		/* 0 == something matched; 1 == nothing matched; 2 or greater == error */
const	char	*srcdir = ".";
const	char	*targetdir = NULL;
const	char	*targetfile = NULL;
const	char	*pattern = NULL;
rule_t	*rule = NULL;


int     main (int argc, char *argv[])
{
	char		str[125+1];
	const	char	*p;
	long	uid;
	int	c;
#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
	int     opt_index;
#endif

	/* set progname */
	progname = *argv;
	if ( (p = strrchr (progname, '/')) )
		progname = ++p;

	/* init global command line parameter vars */

	opterr = 0;
	optind = 1;
#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
	while ( (c = getopt_long (argc, argv, short_options, long_options, &opt_index)) != -1 )
#else
	while ( (c = getopt (argc, argv, short_options)) != -1 )
#endif
	{
		switch ( c )
		{
                case 'c':
			create = 1;
			break;
                case 'e':
			exitmode = 1;
			noexec = 1;
			break;
                case 'f':
			force = 1;
			break;
                case 'l':
			listrules = 1;
			break;
                case 'n':
			noexec = 1;
			break;
                case 'r':
			recursive = 1;
			break;
                case 'v':
			verbose++;
			break;
                case 'V':
			fprintf (stderr, "%s v%s (c) Jan 2015 by hoz\n", progname, VERSION);
			return 0;
                case 'd':
			diff = agefromstr (optarg);
			break;
                case 's':
			srcdir = optarg;
			break;
                case 't':
			targetdir = optarg;
			break;
                case 'T':
			targetfile = optarg;
			singletargetmode = 1;
			break;
                case 'p':
			optionp++;
			pattern = optarg;
			snprintf (str, sizeof (str), "r%d", optionp);
			putrule (str, singletargetmode, pattern);
			break;
#if defined(ROOT_EXEC) && ROOT_EXEC
                case 'X':
			if ( strcmp (optarg, "root-allowed") != 0 )
				usage ("-X requires argument \"root-allowed\"", NULL);
			/* fall trough */
                case 02:
			rootexec = 1;
			break;
#endif
                case 'h':
			usage ("", NULL);
			break;
                case ':':
                        snprintf (str, sizeof(str), "option \"-%c\" requires an argument.\n",
                                                                                optopt);
                        usage (str, NULL);
                        break;
                case '?':
                        if ( isprint (optopt) )
                                snprintf (str, sizeof(str), "Unknown option \"-%c\".\n",
                                                                                optopt);
                        else
                                snprintf (str, sizeof (str), "Unknown option char \\x%x.\n",
                                                                                optopt);
                        usage (str, NULL);
                        break;
                default:
			fprintf (stderr, "Illegal option %c\n", c);
                        abort();
		}
	}

	setup_rules ();

	if ( listrules )
	{
		printrules (stdout);
		exit (2);
	}

	if ( targetdir == NULL )
		targetdir = srcdir;

	c = optind;
	if ( pattern == NULL && c >= argc )	/* no option -p and no arguments left ? */
		fatal ("No rule given\n", NULL);

#ifndef __WIN32__
	if ( (uid = getuid ()) == 0 )
# if defined(ROOT_EXEC) && ROOT_EXEC
		if ( !rootexec )
# endif
			fatal ("execution as user root is not allowed\n", NULL);
#endif
	if ( pattern == NULL )
		rule = getrule (argv[c], singletargetmode);
	else
		rule = getrule ("r1", singletargetmode);
	if ( rule == NULL )
		fatal ("Rule %s not found\n", argv[c]);

	dbg_val ("%s\n", progname);
	dbg_val ("\t-s %s\n", srcdir);
	dbg_val ("\t-t %s\n", targetdir);
	dbg_val3 ("\t:%s:%s:%s:\n", rule->src, rule->target, rule->cmd);

	if ( singletargetmode )
	{
		time_t	t_mtime;
		char	buf[4096];

#if 0
		if ( targetfile == NULL )
			fatal ("option -T used but no target file given\n", NULL);

		if ( fileexist (targetfile) )
#else
		if ( targetfile && fileexist (targetfile) )
#endif
			t_mtime = file_mtime (targetfile);
		else if ( diff )
			t_mtime = time (NULL) - diff;
		else
#if 0
			usage ("option -T used but target file %s does not exist\n", targetfile);
#else
			t_mtime = 0;
#endif
fprintf (stderr, "%ld\n", t_mtime);
		parsedir_sm (buf, sizeof (buf), srcdir, t_mtime, rule);

		if ( verbose )
			fprintf (stderr, "s_list: %s\n", buf);

		if ( buf[0] )	/* source list is not empty */
		{
			if ( exitmode )
				exit (0);
			runcmd (rule->cmd, buf, 0L, targetfile, t_mtime);
		}
	}
	else	/* regular mode */
	{
			/* test if target dir is below src dir */
		if ( recursive && issubdir (srcdir, targetdir) )
			fatal ("target dir (%s) must not be a sub directory of %s in recursive mode\n",
										targetdir, srcdir);
		parsedir_rm (srcdir, targetdir, rule);
		while ( ++c < argc )
		{
			rule = getrule (argv[c], singletargetmode);
			if ( rule == NULL )
				fatal ("Rule %s not found\n", argv[c]);
			parsedir_rm (srcdir, targetdir, rule);
		}
	}

	return exitvalue;
}


#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
# define	lopt_usage(mesg, value) fprintf (stderr, mesg, value)
# define	loptstr(lstr, sstr)     lstr
#else
# define	lopt_usage(mesg, value)
# define	loptstr(lstr, sstr)     sstr
#endif
static	void	usage (const char *mesg, const char *par)
{
	if ( mesg && *mesg )
	{
		fprintf (stderr, mesg, par);
		fprintf (stderr, "\n");
	}

	fprintf (stderr, "usage: %s { -l|-V|-h }\n", progname);
	fprintf (stderr, "\t-l%s", loptstr (", --list\n", ""));
	fprintf (stderr, "\t\t list embedded rules\n");
	fprintf (stderr, "\t-V%s", loptstr (", --version\n", ""));
	fprintf (stderr, "\t\t output version number\n");
	fprintf (stderr, "\t-h%s", loptstr (", --help\n", ""));
	fprintf (stderr, "\t\t display this help\n");
	fprintf (stderr, "\n");

	fprintf (stderr, "usage: %s [-cefnrv] [-s dir] [-t targetdir] [-p <rule>] rulename ...\n", progname);
	fprintf (stderr, "usage: %s [-efnrv]  [-s dir] -T[targetfile] {-p <rule> | rulename }\n", progname);
	fprintf (stderr, "\t-c%s", loptstr (", --create\n", ""));
	fprintf (stderr, "\t\t create target directory if necessary\n");
	fprintf (stderr, "\t-e%s", loptstr (", --exit\n", ""));
	fprintf (stderr, "\t\t don't run a command, just set the exit status\n");
	fprintf (stderr, "\t-f%s", loptstr (", --force\n", ""));
	fprintf (stderr, "\t\t force to run command even if source is older than target\n");
	fprintf (stderr, "\t-n%s", loptstr (", --noexec\n", ""));
	fprintf (stderr, "\t\t do not execute commands\n");
	fprintf (stderr, "\t-r%s", loptstr (", --recursive\n", ""));
	fprintf (stderr, "\t\t run %s in sub directories recursively\n", progname);
	fprintf (stderr, "\t-v%s", loptstr (", --verbose\n", ""));
	fprintf (stderr, "\t\t comment what is done\n");

	fprintf (stderr, "\t-s dir%s", loptstr (", --srcdir=<dir>\n", ""));
	fprintf (stderr, "\t\t specify src directory (default is \".\")\n");
	fprintf (stderr, "\t-t dir%s", loptstr (", --targetdir=<dir>\n", ""));
	fprintf (stderr, "\t\t specify target directory (default is srcdir)\n");
	fprintf (stderr, "\t-T[targetfile]%s", loptstr (", --targetfile[=<file>]\n", ""));
	fprintf (stderr, "\t\t run %s in single target file mode\n", progname);
	fprintf (stderr, "\t-p <pattern>%s", loptstr (", --pattern=<pattern>\n", ""));
	fprintf (stderr, "\t\t specify pattern rule instead of using a builtin one\n");

#if defined(ROOT_EXEC) && ROOT_EXEC
	fprintf (stderr, "\t-X root-allowed%s", loptstr (", --root-exec-allowed\n", "\n\t"));
	fprintf (stderr, "\t\t allow root execution (disabled by default because of security issues)\n");
#endif

	exit (2);
}

typedef struct { const char *name; int single; const char *rule; } brule_t;
static	int	setup_rules (void)
{
	int	i;
	FILE	*fp;
	const	char	*p;
	char		path[MAXPATHLEN];
	static	brule_t	brules[] = {
		{ "cheader",	0,	".h:.c:touch -c -r %sf %tf" },	// this is a trick because of lack of multiple dependency files
		{ "cprog",	0,	".c::cc -o %tf %sf" },		// compile each .c file as separate program

		{ "backup",	0,	":.bak:cp %sf %tf"	},	// create backup files
		{ "copy",	0,	"::cp %sf %tf"	},		// useful if src != target dir
		{ "echo",	0,	"::echo \'%sf --> %tf\'" },
		{ "echo",	1,	"::echo \'%sf --> %tf\'" },
		{ "debug",	0,	"::echo \'%sf (%sd) --> %tf (%td)\'" },
		{ "debug",	1,	"::echo \'%sf --> %tf (%td)\'" },

		{ NULL,		0,	NULL	}
	};

	p = RCFILE;
	if ( (fp = fopen (p, "r")) != NULL )
	{
		readrulefile (fp);
		fclose (fp);
	}
	else if ( verbose >= 2 )
		fprintf (stderr, "Can't open rule file %s\n", p);

	if ( (p = getenv ("HOME")) != NULL )
	{
		snprintf (path, sizeof (path), "%s/%s", p, RCFILE);
		if ( (fp = fopen (path, "r")) != NULL )
		{
			readrulefile (fp);
			fclose (fp);
		}
		else if ( verbose >= 2 )
			fprintf (stderr, "Can't open rule file %s\n", path);
	}

	for ( i = 0; brules[i].name; i++ )
		putrule (brules[i].name, brules[i].single, brules[i].rule);

	return 1;
}

/*****************************************************************
**	parsedir_rm ()	parse directory tree in regular mode
*****************************************************************/
static	int	parsedir_rm (const char *s_dir, const char *t_dir, rule_t *rule)
{
	DIR	*dirp;
	struct  dirent  *dentp;
	char	s_path[MAXPATHLEN+1];
	char	t_path[MAXPATHLEN+1];

	assert ( s_dir != NULL );
	assert ( rule != NULL );

	dbg_val ("directory: opendir(%s)\n", s_dir);
	if ( (dirp = opendir (s_dir)) == NULL )
		return 0;

	/* test if target dir also exist */
	if ( strcmp (s_dir, t_dir) != 0 && !is_directory (t_dir) )
	{
		if ( create == 0 )
			fatal ("target directory %s does not exist: please create it manually or use option -c\n", t_dir);

		if ( verbose )
			fprintf (stderr, "call: mkdir (\"%s\", 0750)\n", t_dir);
		if ( noexec == 0 )
#ifdef __WIN32__
			mkdir (t_dir);
#else
			mkdir (t_dir, 0750);
#endif
	}

	while ( (dentp = readdir (dirp)) != NULL )
	{
		if ( is_dotfilename (dentp->d_name) )
			continue;

		dbg_val ("directory: check %s \n", dentp->d_name);
		pathname (s_path, sizeof (s_path), s_dir, dentp->d_name, NULL, NULL);
		if ( is_directory (s_path) )
		{
			if ( !recursive )	/* skip dirs except in recursive mode */
				continue;
			dbg_val ("directory: recursive %s\n", s_path);
			pathname (t_path, sizeof (t_path), t_dir, dentp->d_name, NULL, NULL);
			parsedir_rm (s_path, t_path, rule);
		}
		else if ( has_fileext (dentp->d_name, rule->src) )
		{
			time_t	s_mtime;
			time_t	t_mtime;

			/* skip all source files with target extension (for empty src extensions) */
			/* but only if the target dir is different */
			if ( *rule->src == '\0' && strcmp (s_dir, t_dir) == 0 &&
						has_fileext (dentp->d_name, rule->target) )
				continue;

			/* this is a candidate to check for a target file */
			if ( verbose >= 2 )
				fprintf (stderr, "%s matches %s\n", dentp->d_name, rule->src);

			/* create target filename */
			pathname (t_path, sizeof (t_path), t_dir, dentp->d_name, rule->src, rule->target);
			if ( strcmp (s_path, t_path) == 0 )	/* Source and target are identical files ? */
			{
				if ( verbose >= 2 )
					fprintf (stderr, "skip identical source and target file\n");

				continue;		/* skip it */
			}

			if ( fileexist (t_path) )
				t_mtime = file_mtime (t_path);
			else
				t_mtime = 0L;

			/* get source timestamp */
			s_mtime = file_mtime (s_path);

			if ( verbose >=2 )
			{
				fprintf (stderr, "%lu %s\n", s_mtime, s_path);
				fprintf (stderr, "%lu %s\n", t_mtime, t_path);
			}
			if ( force || (s_mtime - diff) > t_mtime )	/* source is newer than target ? */
			{
				if ( exitmode )
					exit (0);
				runcmd (rule->cmd, s_path, s_mtime, t_path, t_mtime);
			}
		}
	}
	closedir (dirp);
	return 1;
}


/*****************************************************************
**	parsedir_sm ()	parse directory tree in single target file
**			mode
*****************************************************************/
static	int	parsedir_sm (char *buf, int bufsize, const char *s_dir, time_t ttime, rule_t *rule)
{
	DIR	*dirp;
	struct  dirent  *dentp;
	int	len;
	char	s_path[MAXPATHLEN+1];

	assert ( buf != NULL );
	assert ( s_dir != NULL );
	assert ( rule != NULL );

	dbg_val ("directory: opendir(%s)\n", s_dir);
	if ( (dirp = opendir (s_dir)) == NULL )
		return 0;

	len = 0;
	while ( (dentp = readdir (dirp)) != NULL )
	{
		if ( is_dotfilename (dentp->d_name) )
			continue;

		dbg_val ("directory: check %s \n", dentp->d_name);
		pathname (s_path, sizeof (s_path), s_dir, dentp->d_name, NULL, NULL);
		if ( is_directory (s_path) )
		{
			if ( !recursive )	/* skip dirs except in recursive mode */
				continue;
			dbg_val ("directory: recursive %s\n", s_path);
			len += parsedir_sm (buf+len, bufsize-len, s_path, ttime, rule);
		}
		else if ( has_fileext (dentp->d_name, rule->src) )
		{
			time_t	s_mtime;

			/* this is a candidate to check for a target file */
			if ( verbose >= 2 )
				fprintf (stderr, "%s matches %s\n", dentp->d_name, rule->src);

			/* get source timestamp */
			s_mtime = file_mtime (s_path);

			if ( verbose >=2 )
				fprintf (stderr, "%lu %lu %s\n", ttime, s_mtime, s_path);

			//if ( force || (s_mtime - diff) > ttime )	/* source is newer than target ? */
			if ( force || s_mtime > ttime )	/* source is newer than target ? */
				len += snprintf (buf+len, bufsize-len, "%s ", s_path);
		}
	}
	closedir (dirp);
	return len;
}

/*****************************************************************
**	runcmd ()
**	set the global exitvalue variable to 1
*****************************************************************/
static	int	runcmd (char *cmd, const char *src, time_t s_mtime, const char *tgt, time_t t_mtime)
{
	char	*cmdstr;
	int	ret;

	cmdstr = commandstr (cmd, src, s_mtime, tgt, t_mtime);
	ret = 2;
#if 0
	if ( *cmd == '!' )
	{
		if ( verbose )
			fprintf (stderr, "call: %s\n", cmdstr);
		if ( strncmp (cmd, "!echo ", 5) == 0 )
			echo (src, s_mtime, tgt, t_mtime);
		else if ( strncmp (cmd, "!copy ", 5) == 0 )
			copyfile (src, tgt);
	}
	else
#endif
	{
		if ( verbose )
			fprintf (stderr, "run: %s\n", cmdstr);

		exitvalue = 1;		/* TODO: probably better set if not in noexec mode? */
		if ( noexec != 1 )
		{
#ifdef __WIN32__
			char	*p;

			for ( p = cmdstr; *p; p++ )
				if ( *p == '/' )
					*p = '\\'
#endif
			ret = system (cmdstr);
		}
	}
	return ret;
}

#define	FMTSTR(a, b)	((a) << 8 | (b))
static	char	*commandstr (const char *cmd, const char *src, time_t sdate, const char *tgt, time_t tdate)
{
	static	char	cmdstr[MAXCMDLINE];
	const	char	*s;
	char	*p;
	int	len;

	assert ( cmd != NULL );

	len = 0;
	s = cmd;
	p = cmdstr;
	while ( *s && len < MAXCMDLINE )
	{
		if ( *s == '%' )
		{
			s++;
			if ( *s != '%' )	/* next char is not % ? */
			{
				/* decode format string */
				switch ( FMTSTR(s[0], s[1]) )
				{
				case FMTSTR('s','f'):
					len = snprintf (p, MAXCMDLINE-len, "%s", src);	break;
				case FMTSTR('t','f'):
					len = snprintf (p, MAXCMDLINE-len, "%s", tgt);	break;
				case FMTSTR('s','D'):
					len = snprintf (p, MAXCMDLINE-len, "%lu", sdate);	break;
				case FMTSTR('t','D'):
					len = snprintf (p, MAXCMDLINE-len, "%lu", tdate);	break;
				case FMTSTR('s','d'):
					len = time2isostr (p, MAXCMDLINE-len, sdate, diff);	break;
				case FMTSTR('t','d'):	
					len = time2isostr (p, MAXCMDLINE-len, tdate, diff);	break;
				default:
					fatal ("Error in command string: Illegal format char %%%c%c\n", s[0], s[1]);
				}
				s += 2;
				p += len;
				continue;
			}
		}
		*p++ = *s++;	/* copy char */
		len++;
	}
	*p = '\0';

	str_untaint (cmdstr);
	return cmdstr;
}
